#!/usr/bin/python3

import btrfs
import sys

if len(sys.argv) < 2:
    print("Usage: {} <mountpoint>".format(sys.argv[0]))
    sys.exit(1)


def subvolumes_inside(fs, parent_tree):
    min_key = btrfs.ctree.Key(parent_tree, btrfs.ctree.ROOT_REF_KEY, 0)
    max_key = btrfs.ctree.Key(parent_tree, btrfs.ctree.ROOT_REF_KEY + 1, 0) - 1
    for header, data in btrfs.ioctl.search_v2(fs.fd, 1, min_key, max_key):
        ref = btrfs.ctree.RootRef(header, data)
        path = (btrfs.ioctl.ino_lookup(fs.fd, ref.parent_tree, ref.dirid).name_bytes +
                ref.name).decode()
        yield ref.tree, path


def print_subvolumes_inside(fs, parent_tree, parent_path):
    for tree, path in subvolumes_inside(fs, parent_tree):
        sub_path = "{}/{}".format(parent_path, path)
        print("ID {} parent {} path {}".format(tree, parent_tree, sub_path))
        print_subvolumes_inside(fs, tree, sub_path)


with btrfs.FileSystem(sys.argv[1]) as fs:
    print_subvolumes_inside(fs, 5, '')



#!/usr/bin/python3

import btrfs
import os
import sys

if len(sys.argv) < 6 or len(sys.argv[4:]) % 2 != 0:
    print("Usage: {} <src> <src_offset> <src_length> <dest> <dest_offset> "
          "[<dest> <dest_offset> ...]".format(sys.argv[0]))
    sys.exit(1)

src = sys.argv[1]
src_offset = int(sys.argv[2])
src_length = int(sys.argv[3])

if not os.path.isfile(src):
    print("{} is not a regular file!".format(src))
    sys.exit(1)

dest_names = []
range_infos = []
for n in range(4, len(sys.argv), 2):
    dest = sys.argv[n]
    if not os.path.isfile(dest):
        print("{} is not a regular file!".format(dest))
        sys.exit(1)
    dest_fd = os.open(dest, os.O_RDONLY)
    dest_offset = int(sys.argv[n+1])
    info = btrfs.ioctl.FileDedupeRangeInfo(dest_fd, dest_offset)
    dest_names.append(dest)
    range_infos.append(info)

fd = os.open(src, os.O_RDONLY)
try:
    btrfs.ioctl.fideduperange(fd, src_offset, src_length, range_infos)
    for dest, info in zip(dest_names, range_infos):
        print("dest {} {}".format(dest, info))
finally:
    os.close(fd)
    for info in range_infos:
        os.close(info.dest_fd)



from distutils.core import setup


with open('README.md') as f:
    long_description = f.read()


def get_version():
    version_dict = {}
    with open('btrfs/version.py') as fp:
        exec(fp.read(), version_dict)
    return version_dict['__version__']


version = get_version()

setup(
    name='btrfs',
    packages=['btrfs'],
    version=version,
    description='Python module to interact programmatically with an online btrfs file system',
    long_description=long_description,
    author='Hans van Kranenburg',
    author_email='hans@knorrie.org',
    url='https://github.com/knorrie/python-btrfs',
    download_url='https://github.com/knorrie/python-btrfs/tarball/v{}'.format(version),
    keywords=['btrfs', 'filesystem'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: POSIX :: Linux',
        'Topic :: System :: Filesystems',
        'Topic :: System :: Systems Administration',
    ],
)



#!/usr/bin/python3

import btrfs
import sys

if len(sys.argv) < 2:
    print("Usage: {} <mountpoint>".format(sys.argv[0]))
    sys.exit(1)

with btrfs.FileSystem(sys.argv[1]) as fs:
    fs_info = fs.fs_info()
    print(fs_info)
    for device in fs.devices():
        print(fs.dev_info(device.devid))
        print(fs.dev_stats(device.devid))



#!/usr/bin/python3

import btrfs
import math
import sys

if len(sys.argv) < 2:
    print("Usage: {} [min_score] <mountpoint>".format(sys.argv[0]))
    sys.exit(1)


def extent_tree_free_space_extents(min_vaddr, max_vaddr):
    cur_end = min_vaddr
    for extent in fs.extents(min_vaddr, max_vaddr):
        next_start = extent.vaddr
        if isinstance(extent, btrfs.ctree.ExtentItem):
            next_end = next_start + extent.length
        elif isinstance(extent, btrfs.ctree.MetaDataItem):
            next_end = next_start + fs.nodesize
        if next_start > cur_end:
            yield btrfs.free_space_tree.FreeSpaceExtent(cur_end, next_start - cur_end)
        cur_end = next_end
    if cur_end < max_vaddr:
        yield btrfs.free_space_tree.FreeSpaceExtent(cur_end, max_vaddr + 1 - cur_end)


with btrfs.FileSystem(sys.argv[-1]) as fs:
    if len(sys.argv) == 3:
        min_score = int(sys.argv[1])
    else:
        min_score = 250
    print("Showing all block groups with free space fragmentation score >= {}".format(min_score))

    try:
        list(fs.free_space_extents(0, 0))
        free_space_extents = fs.free_space_extents
    except:
        print("No Free Space Tree (space_cache=v2) found!")
        print("Falling back to using the extent tree to determine free space extents.")
        free_space_extents = extent_tree_free_space_extents

    bad_chunks = 0
    for chunk in fs.chunks():
        if not chunk.type & btrfs.BLOCK_GROUP_DATA:
            continue
        try:
            block_group = fs.block_group(chunk.vaddr, chunk.length)
        except IndexError:
            continue

        min_vaddr = block_group.vaddr
        max_vaddr = block_group.vaddr + block_group.length - 1

        log2_bg_length = math.log2(block_group.length)
        half_width = (log2_bg_length - 11) / 2
        shift = (log2_bg_length + 11) / 2
        # When drawing the function seen below (score += ...), like...
        #
        #         | x - shift  |
        # y = 1 - | ---------- |
        #         | half_width |
        #
        # ... you'll see an upside down V shape. The x values are the log2() of a
        # free space extent size, so e.g. 16 for 64KiB of free space. The y value
        # is the fragmentation score. The function will give a low score to free
        # space fragments that are very small (but when you have an enormous amount
        # of them, it'll add up again) and also a low score to very big ones
        # (they're good). Fragments in the middle get a higher score.
        #
        # Experience learns that scores above 250 point at chunks in which free
        # space fragmentation is getting quite bad. Use btrfs-heatmap to create
        # images of specific block groups to see what's happening inside.
        fragments = 0
        score = 0
        for free_space_extent in free_space_extents(min_vaddr, max_vaddr):
            fragments += 1
            score += 1 - abs((math.log2(free_space_extent.length) - shift) / half_width)
        if score >= min_score:
            bad_chunks += 1
            print("vaddr {} length {} used_pct {} free space fragments {} score {}".format(
                chunk.vaddr, chunk.length, block_group.used_pct, fragments, int(score)))

    if bad_chunks == 0:
        print("No block groups found with free space fragmentation score >= {}".format(min_score))
